@@ -15,7 +15,7 @@
 #'   `minimal`.
 #'
 #' * `unique` names are `minimal`, have no duplicates, and can be used
-#'   where a variable name is expected.  Empty names, and `...` or
+#'   where a variable name is expected. Empty names, `...`, and
 #'   `..` followed by a sequence of digits are banned.
 #'
 #'   - All columns can be accessed by name via `df[["name"]]` and
@@ -77,9 +77,9 @@
 #'
 #' `unique` names are `minimal`, have no duplicates, and can be used
 #'  (possibly with backticks) in contexts where a variable is
-#'  expected. Empty names, and `...` or `..` followed by a sequence of
-#'  digits are banned If a data frame has `unique` names, you can
-#'  index it by name, and also access the columns by name.  In
+#'  expected. Empty names, `...`, and `..` followed by a sequence of
+#'  digits are banned. If a data frame has `unique` names, you can
+#'  index it by name, and also access the columns by name. In
 #'  particular, `df[["name"]]` and `` df$`name` `` and also ``with(df,
 #'  `name`)`` always work.
 #'
@@ -238,10 +238,6 @@ vec_names2 <- function(x,
                        ...,
                        repair = c("minimal", "unique", "universal", "check_unique"),
                        quiet = FALSE) {
-  # This function is an optimized version of:
-  #
-  # vec_as_names(minimal_names(x), ..., repair = repair, quiet = quiet)
-
   if (!missing(...)) {
     ellipsis::check_dots_empty()
   }
@@ -463,11 +459,21 @@ tick_if_needed <- function(x) {
 }
 
 # Used in names.c
-set_rownames <- function(x, names) {
+set_rownames_fallback <- function(x, names) {
   rownames(x) <- names
   x
 }
 
+# Used in names.c
+set_names_fallback <- function(x, names) {
+  names(x) <- names
+  x
+}
+
+vec_set_names <- function(x, names) {
+  .Call(vctrs_set_names, x, names)
+}
+
 #' Repair names with legacy method
 #'
 #' This standardises names with the legacy approach that was used in
