@@ -21,8 +21,22 @@ int32_t ceil2(int32_t x) {
 
 // Dictonary object ------------------------------------------------------------
 
-// Caller is responsible for PROTECTing x
-void dict_init_impl(dictionary* d, SEXP x, bool partial) {
+// Dictionary functions assume that `x` has been proxied recursively because:
+// - `dict_init_impl()` uses `hash_fill()`
+// - `dict_hash_with()` uses `equal_scalar()`
+
+static void dict_init_impl(dictionary* d, SEXP x, bool partial);
+
+// Dictionaries must be protected and unprotected in consistent stack
+// order with `PROTECT_DICT()` and `UNPROTECT_DICT()`.
+void dict_init(dictionary* d, SEXP x) {
+  dict_init_impl(d, x, false);
+}
+void dict_init_partial(dictionary* d, SEXP x) {
+  dict_init_impl(d, x, true);
+}
+
+static void dict_init_impl(dictionary* d, SEXP x, bool partial) {
   d->vec = x;
   d->used = 0;
 
@@ -52,17 +66,6 @@ void dict_init_impl(dictionary* d, SEXP x, bool partial) {
   }
 }
 
-void dict_init(dictionary* d, SEXP x) {
-  dict_init_impl(d, x, false);
-}
-void dict_init_partial(dictionary* d, SEXP x) {
-  dict_init_impl(d, x, true);
-}
-
-void dict_free(dictionary* d) {
-  // no cleanup currently needed
-}
-
 uint32_t dict_hash_with(dictionary* d, dictionary* x, R_len_t i) {
   uint32_t hash = x->hash[i];
 
@@ -110,16 +113,19 @@ void dict_put(dictionary* d, uint32_t hash, R_len_t i) {
 
 SEXP vctrs_unique_loc(SEXP x) {
   int nprot = 0;
+
+  R_len_t n = vec_size(x);
+
   x = PROTECT_N(vec_proxy_equal(x), &nprot);
+  x = PROTECT_N(obj_maybe_translate_encoding(x, n), &nprot);
 
   dictionary d;
   dict_init(&d, x);
+  PROTECT_DICT(&d, &nprot);
 
-  growable g;
-  growable_init(&g, INTSXP, 256);
+  struct growable g = new_growable(INTSXP, 256);
   PROTECT_GROWABLE(&g, &nprot);
 
-  R_len_t n = vec_size(x);
   for (int i = 0; i < n; ++i) {
     uint32_t hash = dict_hash_scalar(&d, i);
 
@@ -131,7 +137,6 @@ SEXP vctrs_unique_loc(SEXP x) {
 
   SEXP out = growable_values(&g);
 
-  dict_free(&d);
   UNPROTECT(nprot);
   return out;
 }
@@ -143,13 +148,18 @@ SEXP vctrs_duplicated_any(SEXP x) {
 
 // [[ include("vctrs.h") ]]
 bool duplicated_any(SEXP x) {
-  x = PROTECT(vec_proxy_equal(x));
+  int nprot = 0;
+
+  R_len_t n = vec_size(x);
+
+  x = PROTECT_N(vec_proxy_equal(x), &nprot);
+  x = PROTECT_N(obj_maybe_translate_encoding(x, n), &nprot);
 
   dictionary d;
   dict_init(&d, x);
+  PROTECT_DICT(&d, &nprot);
 
   bool out = false;
-  R_len_t n = vec_size(x);
 
   for (int i = 0; i < n; ++i) {
     uint32_t hash = dict_hash_scalar(&d, i);
@@ -162,19 +172,22 @@ bool duplicated_any(SEXP x) {
     }
   }
 
-  dict_free(&d);
-  UNPROTECT(1);
-
+  UNPROTECT(nprot);
   return out;
 }
 
 SEXP vctrs_n_distinct(SEXP x) {
-  x = PROTECT(vec_proxy_equal(x));
+  int nprot = 0;
+
+  R_len_t n = vec_size(x);
+
+  x = PROTECT_N(vec_proxy_equal(x), &nprot);
+  x = PROTECT_N(obj_maybe_translate_encoding(x, n), &nprot);
 
   dictionary d;
   dict_init(&d, x);
+  PROTECT_DICT(&d, &nprot);
 
-  R_len_t n = vec_size(x);
   for (int i = 0; i < n; ++i) {
     uint32_t hash = dict_hash_scalar(&d, i);
 
@@ -182,19 +195,23 @@ SEXP vctrs_n_distinct(SEXP x) {
       dict_put(&d, hash, i);
   }
 
-  dict_free(&d);
-  UNPROTECT(1);
+  UNPROTECT(nprot);
   return Rf_ScalarInteger(d.used);
 }
 
 SEXP vctrs_id(SEXP x) {
-  x = PROTECT(vec_proxy_equal(x));
+  int nprot = 0;
+
+  R_len_t n = vec_size(x);
+
+  x = PROTECT_N(vec_proxy_equal(x), &nprot);
+  x = PROTECT_N(obj_maybe_translate_encoding(x, n), &nprot);
 
   dictionary d;
   dict_init(&d, x);
+  PROTECT_DICT(&d, &nprot);
 
-  R_len_t n = vec_size(x);
-  SEXP out = PROTECT(Rf_allocVector(INTSXP, n));
+  SEXP out = PROTECT_N(Rf_allocVector(INTSXP, n), &nprot);
   int* p_out = INTEGER(out);
 
   for (int i = 0; i < n; ++i) {
@@ -206,27 +223,34 @@ SEXP vctrs_id(SEXP x) {
     p_out[i] = d.key[hash] + 1;
   }
 
-  UNPROTECT(2);
-  dict_free(&d);
+  UNPROTECT(nprot);
   return out;
 }
 
 // [[ register() ]]
 SEXP vctrs_match(SEXP needles, SEXP haystack) {
+  int nprot = 0;
   int _;
-  SEXP type = PROTECT(vec_type2(needles, haystack, &args_needles, &args_haystack, &_));
+  SEXP type = PROTECT_N(vec_type2(needles, haystack, &args_needles, &args_haystack, &_), &nprot);
 
-  needles = PROTECT(vec_cast(needles, type, args_empty, args_empty));
-  haystack = PROTECT(vec_cast(haystack, type, args_empty, args_empty));
+  needles = PROTECT_N(vec_cast(needles, type, args_empty, args_empty), &nprot);
+  haystack = PROTECT_N(vec_cast(haystack, type, args_empty, args_empty), &nprot);
 
-  needles = PROTECT(vec_proxy_equal(needles));
-  haystack = PROTECT(vec_proxy_equal(haystack));
+  needles = PROTECT_N(vec_proxy_equal(needles), &nprot);
+  haystack = PROTECT_N(vec_proxy_equal(haystack), &nprot);
+
+  R_len_t n_haystack = vec_size(haystack);
+  R_len_t n_needle = vec_size(needles);
+
+  SEXP translated = PROTECT_N(obj_maybe_translate_encoding2(needles, n_needle, haystack, n_haystack), &nprot);
+  needles = VECTOR_ELT(translated, 0);
+  haystack = VECTOR_ELT(translated, 1);
 
   dictionary d;
   dict_init(&d, haystack);
+  PROTECT_DICT(&d, &nprot);
 
   // Load dictionary with haystack
-  R_len_t n_haystack = vec_size(haystack);
   for (int i = 0; i < n_haystack; ++i) {
     uint32_t hash = dict_hash_scalar(&d, i);
 
@@ -239,8 +263,7 @@ SEXP vctrs_match(SEXP needles, SEXP haystack) {
   dict_init_partial(&d_needles, needles);
 
   // Locate needles
-  R_len_t n_needle = vec_size(needles);
-  SEXP out = PROTECT(Rf_allocVector(INTSXP, n_needle));
+  SEXP out = PROTECT_N(Rf_allocVector(INTSXP, n_needle), &nprot);
   int* p_out = INTEGER(out);
 
   for (int i = 0; i < n_needle; ++i) {
@@ -252,27 +275,35 @@ SEXP vctrs_match(SEXP needles, SEXP haystack) {
     }
   }
 
-  UNPROTECT(6);
-  dict_free(&d);
+  UNPROTECT(nprot);
   return out;
 }
 
 // [[ register() ]]
 SEXP vctrs_in(SEXP needles, SEXP haystack) {
+  int nprot = 0;
+
   int _;
-  SEXP type = PROTECT(vec_type2(needles, haystack, &args_needles, &args_haystack, &_));
+  SEXP type = PROTECT_N(vec_type2(needles, haystack, &args_needles, &args_haystack, &_), &nprot);
 
-  needles = PROTECT(vec_cast(needles, type, args_empty, args_empty));
-  haystack = PROTECT(vec_cast(haystack, type, args_empty, args_empty));
+  needles = PROTECT_N(vec_cast(needles, type, args_empty, args_empty), &nprot);
+  haystack = PROTECT_N(vec_cast(haystack, type, args_empty, args_empty), &nprot);
 
-  needles = PROTECT(vec_proxy_equal(needles));
-  haystack = PROTECT(vec_proxy_equal(haystack));
+  needles = PROTECT_N(vec_proxy_equal(needles), &nprot);
+  haystack = PROTECT_N(vec_proxy_equal(haystack), &nprot);
+
+  R_len_t n_haystack = vec_size(haystack);
+  R_len_t n_needle = vec_size(needles);
+
+  SEXP translated = PROTECT_N(obj_maybe_translate_encoding2(needles, n_needle, haystack, n_haystack), &nprot);
+  needles = VECTOR_ELT(translated, 0);
+  haystack = VECTOR_ELT(translated, 1);
 
   dictionary d;
   dict_init(&d, haystack);
+  PROTECT_DICT(&d, &nprot);
 
   // Load dictionary with haystack
-  R_len_t n_haystack = vec_size(haystack);
   for (int i = 0; i < n_haystack; ++i) {
     uint32_t hash = dict_hash_scalar(&d, i);
 
@@ -283,10 +314,10 @@ SEXP vctrs_in(SEXP needles, SEXP haystack) {
 
   dictionary d_needles;
   dict_init_partial(&d_needles, needles);
+  PROTECT_DICT(&d_needles, &nprot);
 
   // Locate needles
-  R_len_t n_needle = vec_size(needles);
-  SEXP out = PROTECT(Rf_allocVector(LGLSXP, n_needle));
+  SEXP out = PROTECT_N(Rf_allocVector(LGLSXP, n_needle), &nprot);
   int* p_out = LOGICAL(out);
 
   for (int i = 0; i < n_needle; ++i) {
@@ -294,19 +325,25 @@ SEXP vctrs_in(SEXP needles, SEXP haystack) {
     p_out[i] = (d.key[hash] != DICT_EMPTY);
   }
 
-  UNPROTECT(6);
-  dict_free(&d);
+  UNPROTECT(nprot);
   return out;
 }
 
 SEXP vctrs_count(SEXP x) {
-  dictionary d;
-  dict_init(&d, x);
-
-  SEXP val = PROTECT(Rf_allocVector(INTSXP, d.size));
-  int* p_val = INTEGER(val);
+  int nprot = 0;
 
   R_len_t n = vec_size(x);
+
+  x = PROTECT_N(vec_proxy_equal(x), &nprot);
+  x = PROTECT_N(obj_maybe_translate_encoding(x, n), &nprot);
+
+  dictionary d;
+  dict_init(&d, x);
+  PROTECT_DICT(&d, &nprot);
+
+  SEXP val = PROTECT_N(Rf_allocVector(INTSXP, d.size), &nprot);
+  int* p_val = INTEGER(val);
+
   for (int i = 0; i < n; ++i) {
     int32_t hash = dict_hash_scalar(&d, i);
 
@@ -318,8 +355,8 @@ SEXP vctrs_count(SEXP x) {
   }
 
   // Create output
-  SEXP out_key = PROTECT(Rf_allocVector(INTSXP, d.used));
-  SEXP out_val = PROTECT(Rf_allocVector(INTSXP, d.used));
+  SEXP out_key = PROTECT_N(Rf_allocVector(INTSXP, d.used), &nprot);
+  SEXP out_val = PROTECT_N(Rf_allocVector(INTSXP, d.used), &nprot);
   int* p_out_key = INTEGER(out_key);
   int* p_out_val = INTEGER(out_val);
 
@@ -333,29 +370,33 @@ SEXP vctrs_count(SEXP x) {
     i++;
   }
 
-  SEXP out = PROTECT(Rf_allocVector(VECSXP, 2));
+  SEXP out = PROTECT_N(Rf_allocVector(VECSXP, 2), &nprot);
   SET_VECTOR_ELT(out, 0, out_key);
   SET_VECTOR_ELT(out, 1, out_val);
-  SEXP names = PROTECT(Rf_allocVector(STRSXP, 2));
+  SEXP names = PROTECT_N(Rf_allocVector(STRSXP, 2), &nprot);
   SET_STRING_ELT(names, 0, Rf_mkChar("key"));
   SET_STRING_ELT(names, 1, Rf_mkChar("val"));
   Rf_setAttrib(out, R_NamesSymbol, names);
 
-  UNPROTECT(5);
-  dict_free(&d);
+  UNPROTECT(nprot);
   return out;
 }
 
 SEXP vctrs_duplicated(SEXP x) {
-  x = PROTECT(vec_proxy_equal(x));
+  int nprot = 0;
+
+  R_len_t n = vec_size(x);
+
+  x = PROTECT_N(vec_proxy_equal(x), &nprot);
+  x = PROTECT_N(obj_maybe_translate_encoding(x, n), &nprot);
 
   dictionary d;
   dict_init(&d, x);
+  PROTECT_DICT(&d, &nprot);
 
-  SEXP val = PROTECT(Rf_allocVector(INTSXP, d.size));
+  SEXP val = PROTECT_N(Rf_allocVector(INTSXP, d.size), &nprot);
   int* p_val = INTEGER(val);
 
-  R_len_t n = vec_size(x);
   for (int i = 0; i < n; ++i) {
     int32_t hash = dict_hash_scalar(&d, i);
 
@@ -367,7 +408,7 @@ SEXP vctrs_duplicated(SEXP x) {
   }
 
   // Create output
-  SEXP out = PROTECT(Rf_allocVector(LGLSXP, n));
+  SEXP out = PROTECT_N(Rf_allocVector(LGLSXP, n), &nprot);
   int* p_out = LOGICAL(out);
 
   for (int i = 0; i < n; ++i) {
@@ -375,89 +416,7 @@ SEXP vctrs_duplicated(SEXP x) {
     p_out[i] = p_val[hash] != 1;
   }
 
-  UNPROTECT(3);
-  dict_free(&d);
-  return out;
-}
-
-SEXP vctrs_duplicate_split(SEXP x) {
-  x = PROTECT(vec_proxy_equal(x));
-
-  dictionary d;
-  dict_init(&d, x);
-
-  // Tracks the order in which keys are seen
-  SEXP tracker = PROTECT(Rf_allocVector(INTSXP, d.size));
-  int* p_tracker = INTEGER(tracker);
-
-  // Collects the counts of each key
-  SEXP count = PROTECT(Rf_allocVector(INTSXP, d.size));
-  int* p_count = INTEGER(count);
-
-  R_len_t n = vec_size(x);
-
-  // Tells us which element of the index list x[i] goes in
-  SEXP out_pos = PROTECT(Rf_allocVector(INTSXP, n));
-  int* p_out_pos = INTEGER(out_pos);
-
-  // Fill dictionary, out_pos, and count
-  for (int i = 0; i < n; ++i) {
-    uint32_t hash = dict_hash_scalar(&d, i);
-
-    if (d.key[hash] == DICT_EMPTY) {
-      p_tracker[hash] = d.used;
-      dict_put(&d, hash, i);
-      p_count[hash] = 0;
-    }
-
-    p_out_pos[i] = p_tracker[hash];
-    p_count[hash]++;
-  }
-
-  SEXP out_key = PROTECT(Rf_allocVector(INTSXP, d.used));
-  int* p_out_key = INTEGER(out_key);
-
-  SEXP out_idx = PROTECT(Rf_allocVector(VECSXP, d.used));
-
-  SEXP counters = PROTECT(Rf_allocVector(INTSXP, d.used));
-  int* p_counters = INTEGER(counters);
-  memset(p_counters, 0, d.used * sizeof(int));
-
-  // Set up empty index container
-  for (int hash = 0; hash < d.size; ++hash) {
-    if (d.key[hash] == DICT_EMPTY) {
-      continue;
-    }
-
-    SET_VECTOR_ELT(out_idx, p_tracker[hash], Rf_allocVector(INTSXP, p_count[hash]));
-  }
-
-  // Fill index container and key locations
-  for (int i = 0; i < n; ++i) {
-    int j = p_out_pos[i];
-    int hash = p_counters[j];
-
-    if (hash == 0) {
-      p_out_key[j] = i + 1;
-    }
-
-    INTEGER(VECTOR_ELT(out_idx, j))[hash] = i + 1;
-    p_counters[j] = hash + 1;
-  }
-
-  // Construct output
-  SEXP out = PROTECT(Rf_allocVector(VECSXP, 2));
-  SET_VECTOR_ELT(out, 0, out_key);
-  SET_VECTOR_ELT(out, 1, out_idx);
-
-  SEXP names = PROTECT(Rf_allocVector(STRSXP, 2));
-  SET_STRING_ELT(names, 0, Rf_mkChar("key"));
-  SET_STRING_ELT(names, 1, Rf_mkChar("idx"));
-
-  Rf_setAttrib(out, R_NamesSymbol, names);
-
-  UNPROTECT(9);
-  dict_free(&d);
+  UNPROTECT(nprot);
   return out;
 }
 
