@@ -124,6 +124,16 @@ test_that("restore generates correct row/col names", {
   expect_equal(.row_names_info(df2), -2)
 })
 
+test_that("restore keeps automatic row/col names", {
+  df1 <- data.frame(x = NA, y = 1:4, z = 1:4)
+  df1$x <- data.frame(a = 1:4, b = 1:4)
+
+  df2 <- vec_restore(df1, df1)
+
+  expect_named(df2, c("x", "y", "z"))
+  expect_equal(.row_names_info(df2), -4)
+})
+
 test_that("cast to empty data frame preserves number of rows", {
   out <- vec_cast(new_data_frame(n = 10L), new_data_frame())
   expect_equal(nrow(out), 10L)
@@ -134,13 +144,23 @@ test_that("can cast unspecified to data frame", {
   expect_identical(vec_cast(unspecified(3), df), vec_init(df, 3))
 })
 
+test_that("can cast list of data frames to data frame", {
+  df <- data.frame(x = 1, y = 2L)
+  expect_equal(vec_cast(list(df, df), df), vec_slice(df, c(1, 1)))
+})
+
+test_that("can only cast list of data frames to data frame if they are all size 1", {
+  df <- data.frame(x = 1:2)
+  expect_error(vec_cast(list(df), df), class = "vctrs_error_cast_lossy")
+})
+
 test_that("can restore lists with empty names", {
   expect_identical(vec_restore(list(), data.frame()), data.frame())
 })
 
 test_that("can restore subclasses of data frames", {
   expect_identical(vec_restore(list(), subclass(data.frame())), subclass(data.frame()))
-  scoped_global_bindings(
+  local_methods(
     vec_restore.vctrs_foobar = function(x, to, ..., i) "dispatched"
   )
   expect_identical(vec_restore(list(), subclass(data.frame())), "dispatched")
